@@ -65,7 +65,8 @@ enum _GAtServerRequestType {
 
 typedef enum _GAtServerRequestType GAtServerRequestType;
 
-typedef void (*GAtServerNotifyFunc)(GAtServerRequestType type,
+typedef void (*GAtServerNotifyFunc)(GAtServer *server,
+					GAtServerRequestType type,
 					GAtResult *result, gpointer user_data);
 
 GAtServer *g_at_server_new(GIOChannel *io);
@@ -79,6 +80,7 @@ void g_at_server_unref(GAtServer *server);
 
 gboolean g_at_server_shutdown(GAtServer *server);
 
+gboolean g_at_server_set_echo(GAtServer *server, gboolean echo);
 gboolean g_at_server_set_disconnect_function(GAtServer *server,
 					GAtDisconnectFunc disconnect,
 					gpointer user_data);
@@ -86,7 +88,7 @@ gboolean g_at_server_set_debug(GAtServer *server,
 					GAtDebugFunc func,
 					gpointer user_data);
 
-gboolean g_at_server_register(GAtServer *server, char *prefix,
+gboolean g_at_server_register(GAtServer *server, const char *prefix,
 					GAtServerNotifyFunc notify,
 					gpointer user_data,
 					GDestroyNotify destroy_notify);
