@@ -43,14 +43,16 @@ static GHashTable *uuid_hash = NULL;
 static GHashTable *adapter_address_hash = NULL;
 static gint bluetooth_refcount;
 static GSList *server_list = NULL;
+static const char *adapter_any_name = "any";
+static char *adapter_any_path;
 
 #define TIMEOUT 60 /* Timeout for user response (seconds) */
 
 struct server {
 	guint8 channel;
 	char *sdp_record;
+	guint32 handle;
 	GIOChannel *io;
-	GHashTable *adapter_hash;
 	ConnectFunc connect_cb;
 	gpointer user_data;
 };
@@ -218,7 +220,7 @@ void bluetooth_parse_properties(DBusMessage *reply, const char *property, ...)
 	}
 
 done:
-	g_slist_foreach(prop_handlers, (GFunc)g_free, NULL);
+	g_slist_foreach(prop_handlers, (GFunc) g_free, NULL);
 	g_slist_free(prop_handlers);
 }
 
@@ -265,21 +267,20 @@ static void device_properties_cb(DBusPendingCall *call, gpointer user_data)
 	const char *device_addr = NULL;
 	const char *alias = NULL;
 	struct bluetooth_profile *profile;
+	struct DBusError derr;
 
 	reply = dbus_pending_call_steal_reply(call);
 
-	if (dbus_message_is_error(reply, DBUS_ERROR_SERVICE_UNKNOWN)) {
-		DBG("Bluetooth daemon is apparently not available.");
+	dbus_error_init(&derr);
+
+	if (dbus_set_error_from_message(&derr, reply)) {
+		ofono_error("Device.GetProperties replied an error: %s, %s",
+					derr.name, derr.message);
+		dbus_error_free(&derr);
 		goto done;
 	}
 
-	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
-		if (!dbus_message_is_error(reply, DBUS_ERROR_UNKNOWN_METHOD))
-			ofono_info("Error from GetProperties reply: %s",
-					dbus_message_get_error_name(reply));
-
-		goto done;
-	}
+	DBG("");
 
 	bluetooth_parse_properties(reply, "UUIDs", has_uuid, &have_uuid,
 				"Adapter", parse_string, &adapter,
@@ -395,17 +396,24 @@ static void adapter_properties_cb(DBusPendingCall *call, gpointer user_data)
 {
 	const char *path = user_data;
 	DBusMessage *reply;
+	DBusError derr;
 	GSList *device_list = NULL;
 	GSList *l;
 	const char *addr;
 
 	reply = dbus_pending_call_steal_reply(call);
 
-	if (dbus_message_is_error(reply, DBUS_ERROR_SERVICE_UNKNOWN)) {
-		DBG("Bluetooth daemon is apparently not available.");
+	dbus_error_init(&derr);
+
+	if (dbus_set_error_from_message(&derr, reply)) {
+		ofono_error("Adapter.GetProperties replied an error: %s, %s",
+					derr.name, derr.message);
+		dbus_error_free(&derr);
 		goto done;
 	}
 
+	DBG("");
+
 	bluetooth_parse_properties(reply,
 					"Devices", parse_devices, &device_list,
 					"Address", parse_string, &addr,
@@ -437,11 +445,14 @@ static void get_adapter_properties(const char *path, const char *handle,
 			g_strdup(path), g_free, -1, DBUS_TYPE_INVALID);
 }
 
-static void remove_record(char *path, guint handle, struct server *server)
+static void remove_record(struct server *server)
 {
 	DBusMessage *msg;
 
-	msg = dbus_message_new_method_call(BLUEZ_SERVICE, path,
+	if (server->handle == 0)
+		return;
+
+	msg = dbus_message_new_method_call(BLUEZ_SERVICE, adapter_any_path,
 					BLUEZ_SERVICE_INTERFACE,
 					"RemoveRecord");
 	if (msg == NULL) {
@@ -449,24 +460,12 @@ static void remove_record(char *path, guint handle, struct server *server)
 		return;
 	}
 
-	dbus_message_append_args(msg, DBUS_TYPE_UINT32, &handle,
+	dbus_message_append_args(msg, DBUS_TYPE_UINT32, &server->handle,
 					DBUS_TYPE_INVALID);
 	g_dbus_send_message(connection, msg);
 
-	ofono_info("Unregistered handle for %s, channel %d: 0x%x", path,
-			server->channel, handle);
-}
-
-static void server_stop(struct server *server)
-{
-	g_hash_table_foreach_remove(server->adapter_hash,
-					(GHRFunc) remove_record, server);
-
-	if (server->io != NULL) {
-		g_io_channel_shutdown(server->io, TRUE, NULL);
-		g_io_channel_unref(server->io);
-		server->io = NULL;
-	}
+	ofono_info("Unregistered handle for channel %d: 0x%x",
+			server->channel, server->handle);
 }
 
 static void cb_data_destroy(gpointer data)
@@ -542,7 +541,6 @@ static void new_connection(GIOChannel *io, gpointer user_data)
 {
 	struct server *server = user_data;
 	struct cb_data *cbd;
-	guint handle;
 	const char *addr;
 	GError *err = NULL;
 	char laddress[18], raddress[18];
@@ -550,7 +548,6 @@ static void new_connection(GIOChannel *io, gpointer user_data)
 	GHashTableIter iter;
 	gpointer key, value;
 	const char *path;
-	gpointer handlep;
 
 	bt_io_get(io, BT_IO_RFCOMM, &err, BT_IO_OPT_SOURCE, laddress,
 					BT_IO_OPT_DEST, raddress,
@@ -578,10 +575,6 @@ static void new_connection(GIOChannel *io, gpointer user_data)
 	if (path == NULL)
 		return;
 
-	handlep = g_hash_table_lookup(server->adapter_hash, path);
-	if (handlep == NULL)
-		return;
-
 	cbd = g_try_new0(struct cb_data, 1);
 	if (cbd == NULL) {
 		ofono_error("Unable to allocate client cb_data structure");
@@ -593,30 +586,35 @@ static void new_connection(GIOChannel *io, gpointer user_data)
 	cbd->io = io;
 
 	addr = raddress;
-	handle = GPOINTER_TO_UINT(handlep);
 
 	if (bluetooth_send_with_reply(path, BLUEZ_SERVICE_INTERFACE,
 					"RequestAuthorization",
 					auth_cb, cbd, cb_data_destroy,
 					TIMEOUT, DBUS_TYPE_STRING, &addr,
-					DBUS_TYPE_UINT32, &handle,
+					DBUS_TYPE_UINT32, &server->handle,
 					DBUS_TYPE_INVALID) < 0) {
 		ofono_error("Request Bluetooth authorization failed");
 		return;
 	}
 
-	ofono_info("RequestAuthorization(%s, 0x%x)", raddress, handle);
+	ofono_info("RequestAuthorization(%s, 0x%x)", raddress, server->handle);
 
 	cbd->source = g_io_add_watch(io, G_IO_HUP | G_IO_ERR | G_IO_NVAL,
 					client_event, cbd);
 }
 
+static void remove_service_handle(gpointer data, gpointer user_data)
+{
+	struct server *server = data;
+
+	server->handle = 0;
+}
+
 static void add_record_cb(DBusPendingCall *call, gpointer user_data)
 {
-	struct cb_data *cb_data = user_data;
+	struct server *server = user_data;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	DBusError derr;
-	guint32 handle;
 
 	dbus_error_init(&derr);
 
@@ -624,47 +622,57 @@ static void add_record_cb(DBusPendingCall *call, gpointer user_data)
 		ofono_error("Replied with an error: %s, %s",
 					derr.name, derr.message);
 		dbus_error_free(&derr);
-		g_free(cb_data->path);
 		goto done;
 	}
 
-	dbus_message_get_args(reply, NULL, DBUS_TYPE_UINT32, &handle,
+	dbus_message_get_args(reply, NULL, DBUS_TYPE_UINT32, &server->handle,
 					DBUS_TYPE_INVALID);
 
-	g_hash_table_insert(cb_data->server->adapter_hash, cb_data->path,
-				GUINT_TO_POINTER(handle));
-
-	ofono_info("Registered handle for %s, channel %d: 0x%x", cb_data->path,
-			cb_data->server->channel, handle);
+	ofono_info("Registered handle for channel %d: 0x%x",
+			server->channel, server->handle);
 
 done:
-	/* Do not free cb_data->path, it is used in adapter_hash */
-	g_free(cb_data);
 	dbus_message_unref(reply);
 }
 
 static void add_record(gpointer data, gpointer user_data)
 {
 	struct server *server = data;
-	const char *path = user_data;
-	struct cb_data *cb_data;
 
 	if (server->sdp_record == NULL)
 		return;
 
-	cb_data = g_try_new0(struct cb_data, 1);
-	if (cb_data == NULL) {
-		ofono_error("Unable to allocate cb_data structure");
-		return;
+	bluetooth_send_with_reply(adapter_any_path,
+					BLUEZ_SERVICE_INTERFACE, "AddRecord",
+					add_record_cb, server, NULL, -1,
+					DBUS_TYPE_STRING, &server->sdp_record,
+					DBUS_TYPE_INVALID);
+}
+
+static void find_adapter_cb(DBusPendingCall *call, gpointer user_data)
+{
+	DBusMessage *reply = dbus_pending_call_steal_reply(call);
+	DBusError derr;
+	const char *path;
+
+	dbus_error_init(&derr);
+
+	if (dbus_set_error_from_message(&derr, reply)) {
+		ofono_error("Replied with an error: %s, %s",
+					derr.name, derr.message);
+		dbus_error_free(&derr);
+		goto done;
 	}
 
-	cb_data->server = server;
-	cb_data->path = g_strdup(path);
+	dbus_message_get_args(reply, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+					DBUS_TYPE_INVALID);
 
-	bluetooth_send_with_reply(path, BLUEZ_SERVICE_INTERFACE, "AddRecord",
-				add_record_cb, cb_data, NULL, -1,
-				DBUS_TYPE_STRING, &server->sdp_record,
-				DBUS_TYPE_INVALID);
+	adapter_any_path = g_strdup(path);
+
+	g_slist_foreach(server_list, (GFunc) add_record, NULL);
+
+done:
+	dbus_message_unref(reply);
 }
 
 static gboolean adapter_added(DBusConnection *connection, DBusMessage *message,
@@ -680,8 +688,6 @@ static gboolean adapter_added(DBusConnection *connection, DBusMessage *message,
 			"GetProperties", adapter_properties_cb, g_strdup(path),
 			g_free, -1, DBUS_TYPE_INVALID);
 
-	g_slist_foreach(server_list, add_record, (gpointer) path);
-
 	return TRUE;
 }
 
@@ -689,19 +695,11 @@ static gboolean adapter_removed(DBusConnection *connection,
 				DBusMessage *message, void *user_data)
 {
 	const char *path;
-	GSList *l;
 
 	if (dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
 				DBUS_TYPE_INVALID) == TRUE)
 		g_hash_table_remove(adapter_address_hash, path);
 
-	for (l = server_list; l; l = l->next) {
-		struct server *server = l->data;
-
-		/* Handle have already been removed, so removing related path */
-		g_hash_table_remove(server->adapter_hash, path);
-	}
-
 	return TRUE;
 }
 
@@ -728,8 +726,6 @@ static void parse_adapters(DBusMessageIter *array, gpointer user_data)
 				"GetProperties", adapter_properties_cb,
 				g_strdup(path), g_free, -1, DBUS_TYPE_INVALID);
 
-		g_slist_foreach(server_list, add_record, (gpointer) path);
-
 		dbus_message_iter_next(&value);
 	}
 }
@@ -737,11 +733,16 @@ static void parse_adapters(DBusMessageIter *array, gpointer user_data)
 static void manager_properties_cb(DBusPendingCall *call, gpointer user_data)
 {
 	DBusMessage *reply;
+	DBusError derr;
 
 	reply = dbus_pending_call_steal_reply(call);
 
-	if (dbus_message_is_error(reply, DBUS_ERROR_SERVICE_UNKNOWN)) {
-		DBG("Bluetooth daemon is apparently not available.");
+	dbus_error_init(&derr);
+
+	if (dbus_set_error_from_message(&derr, reply)) {
+		ofono_error("Manager.GetProperties() replied an error: %s, %s",
+					derr.name, derr.message);
+		dbus_error_free(&derr);
 		goto done;
 	}
 
@@ -762,12 +763,26 @@ static void bluetooth_remove_all_modem(gpointer key, gpointer value,
 	profile->remove_all();
 }
 
+static void bluetooth_connect(DBusConnection *connection, void *user_data)
+{
+	bluetooth_send_with_reply("/", BLUEZ_MANAGER_INTERFACE, "GetProperties",
+				manager_properties_cb, NULL, NULL, -1,
+				DBUS_TYPE_INVALID);
+
+	bluetooth_send_with_reply("/", BLUEZ_MANAGER_INTERFACE, "FindAdapter",
+				find_adapter_cb, NULL, NULL, -1,
+				DBUS_TYPE_STRING, &adapter_any_name,
+				DBUS_TYPE_INVALID);
+}
+
 static void bluetooth_disconnect(DBusConnection *connection, void *user_data)
 {
 	if (uuid_hash == NULL)
 		return;
 
 	g_hash_table_foreach(uuid_hash, bluetooth_remove_all_modem, NULL);
+
+	g_slist_foreach(server_list, (GFunc) remove_service_handle, NULL);
 }
 
 static guint bluetooth_watch;
@@ -783,7 +798,8 @@ static void bluetooth_ref(void)
 	connection = ofono_dbus_get_connection();
 
 	bluetooth_watch = g_dbus_add_service_watch(connection, BLUEZ_SERVICE,
-					NULL, bluetooth_disconnect, NULL, NULL);
+					bluetooth_connect,
+					bluetooth_disconnect, NULL, NULL);
 
 	adapter_added_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
 						BLUEZ_MANAGER_INTERFACE,
@@ -811,10 +827,6 @@ static void bluetooth_ref(void)
 	adapter_address_hash = g_hash_table_new_full(g_str_hash, g_str_equal,
 						g_free, g_free);
 
-	bluetooth_send_with_reply("/", BLUEZ_MANAGER_INTERFACE, "GetProperties",
-				manager_properties_cb, NULL, NULL, -1,
-				DBUS_TYPE_INVALID);
-
 increment:
 	g_atomic_int_inc(&bluetooth_refcount);
 
@@ -832,6 +844,9 @@ static void bluetooth_unref(void)
 	if (g_atomic_int_dec_and_test(&bluetooth_refcount) == FALSE)
 		return;
 
+	g_free(adapter_any_path);
+	adapter_any_path = NULL;
+
 	g_dbus_remove_watch(connection, bluetooth_watch);
 	g_dbus_remove_watch(connection, adapter_added_watch);
 	g_dbus_remove_watch(connection, adapter_removed_watch);
@@ -865,8 +880,6 @@ struct server *bluetooth_register_server(guint8 channel, const char *sdp_record,
 {
 	struct server *server;
 	GError *err;
-	GHashTableIter iter;
-	gpointer key, value;
 
 	server = g_try_new0(struct server, 1);
 	if (!server)
@@ -892,24 +905,27 @@ struct server *bluetooth_register_server(guint8 channel, const char *sdp_record,
 
 	server->connect_cb = cb;
 	server->user_data = user_data;
-	server->adapter_hash = g_hash_table_new_full(g_str_hash, g_str_equal,
-							g_free, NULL);
-
-	g_hash_table_iter_init(&iter, adapter_address_hash);
-
-	while (g_hash_table_iter_next(&iter, &key, &value))
-		add_record(server, key);
 
 	server_list = g_slist_prepend(server_list, server);
 
+	if (adapter_any_path != NULL)
+		add_record(server, NULL);
+
 	return server;
 }
 
 void bluetooth_unregister_server(struct server *server)
 {
 	server_list = g_slist_remove(server_list, server);
-	server_stop(server);
-	g_hash_table_destroy(server->adapter_hash);
+
+	remove_record(server);
+
+	if (server->io != NULL) {
+		g_io_channel_shutdown(server->io, TRUE, NULL);
+		g_io_channel_unref(server->io);
+		server->io = NULL;
+	}
+
 	g_free(server->sdp_record);
 	g_free(server);
 
